// file      : xsd-frontend/parser.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <xsd-frontend/parser.hxx>

#include <xsd-frontend/types.hxx>
#include <xsd-frontend/xml.hxx>
#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/map.hxx>
#include <cult/containers/stack.hxx>

//@@ Do i need this?
//
#include <xercesc/dom/DOM.hpp>

#include <xercesc/sax/ErrorHandler.hpp>
#include <xercesc/sax/SAXParseException.hpp>

#include <xercesc/sax2/SAX2XMLReader.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>

#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/BinInputStream.hpp>
#include <xercesc/util/BinFileInputStream.hpp>

#include <xercesc/validators/common/Grammar.hpp>
#include <xercesc/validators/schema/XSDDOMParser.hpp>

#include <xercesc/sax/InputSource.hpp>
#include <xercesc/framework/LocalFileInputSource.hpp>
#include <xercesc/framework/Wrapper4InputSource.hpp>  //@@ Need this?
#include <xercesc/framework/Wrapper4DOMInputSource.hpp>  //@@ Need this?

#include <cstdlib>  // std::wcstombs
#include <iostream>
#include <memory>   // std::auto_ptr

using std::wcout;
using std::wcerr;
using std::endl;

namespace XSDFrontend
{
  namespace Xerces = XML::Xerces;
  using namespace SemanticGraph;

  //@@ Port to tracing facility.
  //
  Boolean trace_ = false;

  String const xsd = L"http://www.w3.org/2001/XMLSchema";
  String const xse = L"http://www.codesynthesis.com/xmlns/xml-schema-extension";

  namespace
  {
    // Exceptions.
    //

    struct NotNamespace
    {
      NotNamespace (String const& ns)
          : ns_ (ns)
      {
      }

      String const&
      ns () const
      {
        return ns_;
      }

    private:
      String ns_;
    };

    struct NotName
    {
      NotName (String const& ns, String const& name)
          : ns_ (ns), name_ (name)
      {
      }

      String const&
      ns () const
      {
        return ns_;
      }

      String const&
      name () const
      {
        return name_;
      }

    private:
      String ns_;
      String name_;
    };


    template <typename X>
    X&
    resolve (String const& ns_name,
             String const& uq_name,
             Schema& s_)
    {
      Scope::NamesIteratorPair nss (s_.find (ns_name));

      if (nss.first == nss.second)
        throw NotNamespace (ns_name);

      for (; nss.first != nss.second; ++nss.first)
      {
        Namespace& ns (dynamic_cast<Namespace&> (nss.first->named ()));

        Scope::NamesIteratorPair types (ns.find (uq_name));

        for (; types.first != types.second; ++types.first)
        {
          if (X* x = dynamic_cast<X*> (&types.first->named ()))
          {
            if (trace_)
              wcout << "successfully resolved '" << ns_name << '#' << uq_name
                    << "'" << endl;

            return *x;
          }
        }
      }

      throw NotName (ns_name, uq_name);
    }

    struct Resolver : Traversal::Member,
                      Traversal::Element,
                      Traversal::Fundamental::IdRef,
                      Traversal::Fundamental::IdRefs,
                      Traversal::List,
                      Traversal::Complex,
                      Traversal::Enumeration,
                      Traversal::Scope // group, attributeGroup
    {
      Resolver (Schema& s, Boolean& valid)
          : s_ (s), valid_ (valid), may_need_more_ (false)
      {
      }

      Boolean
      need_more ()
      {
        return may_need_more_;
      }

      Void
      traverse (SemanticGraph::Element& e)
      {
        {
          SemanticGraph::Member& m (e);
          traverse (m);
        }

        if (e.context ().count ("substitution-ns-name"))
        {
          String ns_name (e.context ().get<String> ("substitution-ns-name"));
          String uq_name (e.context ().get<String> ("substitution-uq-name"));

          e.context ().remove ("substitution-ns-name");
          e.context ().remove ("substitution-uq-name");

          try
          {
            SemanticGraph::Element& root (
              resolve<SemanticGraph::Element> (ns_name, uq_name, s_));

            s_.new_edge<Substitutes> (e, root);
          }
          catch (NotNamespace const& ex)
          {
            if (valid_)
            {
              wcerr << "ice: unable to resolve namespace '" << ex.ns () << "'"
                    << endl;

              abort ();
            }
          }
          catch (NotName const& ex)
          {
            if (valid_)
            {
              wcerr << "ice: unable to resolve name '" << ex.name ()
                    << "' inside namespace '" << ex.ns () << "'" <<endl;

              abort ();
            }
          }
        }

      }

      Void
      traverse (SemanticGraph::Member& i)
      {
        try
        {
          String ns_name;
          String uq_name;

          if (i.context ().count ("type-ns-name"))
          {
            ns_name = i.context ().get<String> ("type-ns-name");
            uq_name = i.context ().get<String> ("type-uq-name");

            i.context ().remove ("type-ns-name");
            i.context ().remove ("type-uq-name");

            s_.new_edge<Belongs> (
              i, resolve<SemanticGraph::Type> (ns_name, uq_name, s_));
          }
          else if (i.context ().count ("instance-ns-name"))
          {
            ns_name = i.context ().get<String> ("instance-ns-name");
            uq_name = i.context ().get<String> ("instance-uq-name");

            i.context ().remove ("instance-ns-name");
            i.context ().remove ("instance-uq-name");

            SemanticGraph::Member& ref (
              resolve<SemanticGraph::Member> (ns_name, uq_name, s_));

            s_.new_edge<BelongsToNamespace> (i, ref.namespace_ ());

            if (ref.typed ())
            {
              s_.new_edge<Belongs> (i, ref.type ());
            }
            else if (valid_)
            {
              wcerr << "ice: referenced instance '" << ns_name << "#"
                    << uq_name << "' is not typed" << endl;

              abort ();
            }
          }
        }
        catch (NotNamespace const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve namespace '" << ex.ns () << "'"
                  << endl;

            abort ();
          }
        }
        catch (NotName const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve name '" << ex.name ()
                  << "' inside namespace '" << ex.ns () << "'" <<endl;

            abort ();
          }
        }
      }

      Void
      traverse (SemanticGraph::Fundamental::IdRef& i)
      {
        ref_type (i);
      }

      Void
      traverse (SemanticGraph::Fundamental::IdRefs& i)
      {
        ref_type (i);
      }

      Void
      ref_type (SemanticGraph::Specialization& s)
      {
        if (s.context ().count ("type-ns-name"))
        {
          String ns_name = s.context ().get<String> ("type-ns-name");
          String uq_name = s.context ().get<String> ("type-uq-name");

          s.context ().remove ("type-ns-name");
          s.context ().remove ("type-uq-name");

          try
          {
            s_.new_edge<Arguments> (
              resolve<SemanticGraph::Type> (ns_name, uq_name, s_), s);
          }
          catch (NotName const& ex)
          {
            wcerr << s.file () << ":" << s.line () << ":" << s.column () << ": "
                  << "error: unable to resolve type '" << uq_name << "' "
                  << "in namespace '" << ns_name << "'" << endl;

            valid_ = false;
          }
        }
      }


      Void
      traverse (SemanticGraph::List& l)
      {
        if (l.context ().count ("type-ns-name"))
        {
          String ns_name = l.context ().get<String> ("type-ns-name");
          String uq_name = l.context ().get<String> ("type-uq-name");

          l.context ().remove ("type-ns-name");
          l.context ().remove ("type-uq-name");

          s_.new_edge<Arguments> (
            resolve<SemanticGraph::Type> (ns_name, uq_name, s_), l);
        }
      }

      Void
      traverse (SemanticGraph::Complex& c)
      {
        if (c.context ().count ("type-ns-name"))
        {
          String ns_name = c.context ().get<String> ("type-ns-name");
          String uq_name = c.context ().get<String> ("type-uq-name");

          c.context ().remove ("type-ns-name");
          c.context ().remove ("type-uq-name");

          s_.new_edge<Inherits> (
            c, resolve<SemanticGraph::Type> (ns_name, uq_name, s_));
        }

        Traversal::Complex::traverse (c);
      }

      Void
      traverse (SemanticGraph::Scope& s)
      {
        if (s.context ().count ("group-ns-name"))
        {
          traverse_group (s);
        }
        else if (s.context ().count ("attrgroup-ns-name"))
        {
          traverse_attribute_group (s);
        }
        else
          Traversal::Scope::traverse (s);
      }


      Void
      traverse_group (SemanticGraph::Scope& s)
      {
        using SemanticGraph::Element;

        SemanticGraph::Scope& parent (s.scope ());

        String ns_name (s.context ().get<String> ("group-ns-name"));
        String uq_name (s.context ().get<String> ("group-uq-name"));

        UnsignedLong min (s.context ().get<UnsignedLong> ("group-min"));
        UnsignedLong max (s.context ().get<UnsignedLong> ("group-max"));

        // Need to remove the scope altogether. For now I just remove
        // stuff from the context so it won't trigger on the subsequent
        // resolution passes.
        //
        s.context ().remove ("group-ns-name");
        s.context ().remove ("group-uq-name");
        s.context ().remove ("group-min");
        s.context ().remove ("group-max");

        try
        {
          Scope& s (resolve<SemanticGraph::Scope> (ns_name, uq_name, s_));

          for (Scope::NamesIterator i (s.names_begin ());
               i != s.names_end (); ++i)
          {
            // Normally a referenced Group would be replaced with its
            // content. However it is not possible when the Group hasn't
            // been defined yet. In this case we simply copy the group
            // here and let resolver handle this.
            //
            if (dynamic_cast<Scope*> (&(i->named ())))
            {
              Scope& p (dynamic_cast<Scope&> (i->named ()));

              if (!p.context ().count ("group-ns-name"))
                continue;

              Scope& ref (
                s_.new_node<Scope> (p.file (), p.line (), p.column ()));

              String ns_name (p.context ().get<String> ("group-ns-name"));
              String uq_name (p.context ().get<String> ("group-uq-name"));
              UnsignedLong pmin (p.context ().get<UnsignedLong> ("group-min"));
              UnsignedLong pmax (p.context ().get<UnsignedLong> ("group-max"));

              ref.context ().set ("group-ns-name", ns_name);
              ref.context ().set ("group-uq-name", uq_name);
              ref.context ().set ("group-min", pmin == 0 ? 0 : min);
              ref.context ().set ("group-max", pmax != 1 ? pmax : max);

              s_.new_edge<Names> (parent, ref, L"");

              may_need_more_ = true;
            }
            else
            {
              // Prototype.
              //
              Element& p (dynamic_cast<Element&> (i->named ()));

              Element& e (
                s_.new_node<Element> (p.file (),
                                      p.line (),
                                      p.column (),
                                      min == 0 ? min : p.min (),
                                      max == 0 ? max : p.max (),
                                      p.global (),
                                      p.qualified ()));

              s_.new_edge<Names> (parent, e, p.name ());

              if (p.qualified () && !p.context ().count ("instance-ns-name"))
                s_.new_edge<BelongsToNamespace> (e, p.namespace_ ());

              // Now we set element's type.
              //
              if (p.typed ())
              {
                // Easy case.
                //
                s_.new_edge<Belongs> (e, p.type ());
              }
              else if (p.context ().count ("type-ns-name"))
              {
                String ns_name (p.context ().get<String> ("type-ns-name"));
                String uq_name (p.context ().get<String> ("type-uq-name"));

                e.context ().set ("type-ns-name", ns_name);
                e.context ().set ("type-uq-name", uq_name);

                may_need_more_ = true;
              }
              else if (p.context ().count ("instance-ns-name"))
              {
                String ns_name (p.context ().get<String> ("instance-ns-name"));
                String uq_name (p.context ().get<String> ("instance-uq-name"));

                e.context ().set ("instance-ns-name", ns_name);
                e.context ().set ("instance-uq-name", uq_name);

                may_need_more_ = true;
              }
              else if (valid_)
              {
                wcerr << "ice: element '" << ns_name << "#" << uq_name << "' "
                      << "is in unexpected condition" << endl;

                abort ();
              }
            }
          }
        }
        catch (NotNamespace const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve namespace '" << ex.ns () << "'"
                  << endl;

            abort ();
          }
        }
        catch (NotName const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve name '" << ex.name ()
                  << "' inside namespace '" << ex.ns () << "'" << endl;

            abort ();
          }
        }
      }


      Void
      traverse_attribute_group (SemanticGraph::Scope& s)
      {
        SemanticGraph::Scope& parent (s.scope ());

        String ns_name (s.context ().get<String> ("attrgroup-ns-name"));
        String uq_name (s.context ().get<String> ("attrgroup-uq-name"));

        // Need to remove the scope altogether. For now I just remove
        // stuff from the context so it won't trigger on the subsequent
        // resolution passes.
        //
        s.context ().remove ("attrgroup-ns-name");
        s.context ().remove ("attrgroup-uq-name");

        try
        {
          Scope& s (resolve<Scope> (ns_name, uq_name, s_));

          for (Scope::NamesIterator i (s.names_begin ());
               i != s.names_end (); ++i)
          {
            // Normally a referenced attributeGroup would be replaced
            // with its content. However it is not possible when
            // attributeGroup hasn't been defined yet. In this case
            // we simply copy the group here and let resolver handle
            // this.
            //
            if (dynamic_cast<Scope*> (&(i->named ())))
            {
              Scope& p (dynamic_cast<Scope&> (i->named ()));

              if (!p.context ().count ("attrgroup-ns-name"))
                continue;

              Scope& ref (
                s_.new_node<Scope> (p.file (), p.line (), p.column ()));

              String ns_name (p.context ().get<String> ("attrgroup-ns-name"));
              String uq_name (p.context ().get<String> ("attrgroup-uq-name"));

              ref.context ().set ("attrgroup-ns-name", ns_name);
              ref.context ().set ("attrgroup-uq-name", uq_name);

              s_.new_edge<Names> (parent, ref, L"");

              may_need_more_ = true;
            }
            else
            {
              // Prototype.
              //
              Attribute& p (dynamic_cast<Attribute&> (i->named ()));

              Attribute& a (
                s_.new_node<Attribute> (p.file (),
                                        p.line (),
                                        p.column (),
                                        p.optional (),
                                        p.global (),
                                        p.qualified ()));


              s_.new_edge<Names> (parent, a, p.name ());

              if (p.qualified () && !p.context ().count ("instance-ns-name"))
                s_.new_edge<BelongsToNamespace> (a, p.namespace_ ());

              // Set attribute type.
              //
              if (p.typed ())
              {
                // Easy case.
                //
                s_.new_edge<Belongs> (a, p.type ());
              }
              else if (p.context ().count ("type-ns-name"))
              {
                String ns_name (p.context ().get<String> ("type-ns-name"));
                String uq_name (p.context ().get<String> ("type-uq-name"));

                a.context ().set ("type-ns-name", ns_name);
                a.context ().set ("type-uq-name", uq_name);

                may_need_more_ = true;
              }
              else if (p.context ().count ("instance-ns-name"))
              {
                String ns_name (p.context ().get<String> ("instance-ns-name"));
                String uq_name (p.context ().get<String> ("instance-uq-name"));

                a.context ().set ("instance-ns-name", ns_name);
                a.context ().set ("instance-uq-name", uq_name);

                may_need_more_ = true;
              }
              else if (valid_)
              {
                wcerr << "ice: attribute is in unexpected condition" << endl;

                abort ();
              }
            }
          }
        }
        catch (NotNamespace const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve namespace '" << ex.ns () << "'"
                  << endl;

            abort ();
          }
        }
        catch (NotName const& ex)
        {
          if (valid_)
          {
            wcerr << "ice: unable to resolve attribute group name '"
                  << ex.name () << "' inside namespace '" << ex.ns () << "'"
                  << endl;

            abort ();
          }
        }
      }

      // We don't need to go inside enumerations.
      //
      Void
      traverse (SemanticGraph::Enumeration&)
      {
      }

    private:
      Schema& s_;
      Boolean& valid_;
      Boolean may_need_more_;
    };
  }

  //
  //
  struct FilePathComparator
  {
    Boolean
    operator () (SemanticGraph::Path const& x,
                 SemanticGraph::Path const& y) const
    {
        return x.string () < y.string ();
    }
  };

  // Parser::Impl
  //

  class Parser::Impl: public NonCopyable
  {
  public:
    Impl ();

    Evptr<SemanticGraph::Schema>
    parse (SemanticGraph::Path const&);

  private:
    XML::AutoPtr<Xerces::DOMDocument>
    dom (SemanticGraph::Path const&);

    Void
    schema (XML::Element const&);

    Void
    annotation ();

    Void
    import (XML::Element const&);

    Void
    include (XML::Element const&);

    Void
    group (XML::Element const&);

    SemanticGraph::Type*
    simple_type (XML::Element const&);

    SemanticGraph::Type*
    list (XML::Element const& l, XML::Element const& type);

    SemanticGraph::Type*
    restriction (XML::Element const& r, XML::Element const& type);

    Void
    enumeration (XML::Element const&);

    SemanticGraph::Type*
    complex_type (XML::Element const&);

    Void
    all (XML::Element const&);

    Void
    choice (XML::Element const&);

    Void
    sequence (XML::Element const&);

    Void
    simple_content (XML::Element const&);

    Void
    complex_content (XML::Element const&);

    Void
    simple_content_extension (XML::Element const&);

    Void
    simple_content_restriction (XML::Element const&);

    Void
    complex_content_extension (XML::Element const&);

    Void
    complex_content_restriction (XML::Element const&);

    Void
    element (XML::Element const&, Boolean global = false);

    Void
    attribute (XML::Element const&, Boolean global = false);

    Void
    attribute_group (XML::Element const&);

    Void
    any (XML::Element const&);

    Void
    any_attribute (XML::Element const&);

  private:
    Boolean
    more () const
    {
      Iterator const& it (iteration_state_.top ());

      return it.l_->getLength () > it.i_;
    }

    XML::Element
    next ()
    {
      Iterator& it (iteration_state_.top ());

      return XML::Element (
        dynamic_cast<Xerces::DOMElement*> (it.l_->item (it.i_++)));
    }

    Void
    prev ()
    {
      Iterator& it (iteration_state_.top ());

      if (it.i_)
        --it.i_;
    }

    Void
    push (XML::Element const& e)
    {
      iteration_state_.push (e.dom_element ());
    }

    Void
    pop ()
    {
      iteration_state_.pop ();
    }

  private:
    Void
    push_scope (SemanticGraph::Scope& s)
    {
      scope_stack_.push (&s);
    }

    Void
    pop_scope ()
    {
      scope_stack_.pop ();
    }

    SemanticGraph::Scope&
    scope () const
    {
      return *(scope_stack_.top ());
    }

  private:
    Void
    push_cardinality (UnsignedLong min, UnsignedLong max)
    {
      cardinality_stack_.push (Cardinality (min, max));
    }

    Void
    pop_cardinality ()
    {
      cardinality_stack_.pop ();
    }

    UnsignedLong
    min () const
    {
      if (cardinality_stack_.empty ())
        return 1;
      else
        return cardinality_stack_.top ().min_;
    }

    UnsignedLong
    max () const
    {
      if (cardinality_stack_.empty ())
        return 1;
      else
        return cardinality_stack_.top ().max_;
    }

  private:
    SemanticGraph::Namespace&
    cur_ns () const
    {
      // Here I am using the fact that each Schema Names only one
      // Namespace.
      //
      return dynamic_cast<Namespace&> (cur_->names_begin ()->named ());
    }

  private:
    // Some useful tests.
    //

    // Return true if cur_ is a schema with targetNamespace. Note that
    // it returns false if this schema is a chameleon.
    //
    bool
    has_target_ns () const
    {
      if (cur_ns().name ().empty ())
        return false;

      // Test if this is a chameleon schema.
      // @@ What if this schema is both a chameleon and something else?
      //
      if (cur_->contained () &&
          dynamic_cast<Sources*> (&(*cur_->contained_begin ())) != 0)
        return false;

      return true;
    }

  private:
    template <typename Edge, typename Node>
    Void
    set_type (String const& type, XML::Element const& e, Node& node);

  private:
    struct Iterator
    {
      Iterator (Xerces::DOMElement* e)
          : l_ (e->getChildNodes ()), i_ (0)
      {
      }

      Xerces::DOMNodeList* l_;
      UnsignedLong i_;
    };

    Cult::Containers::Stack<Iterator> iteration_state_;
    SemanticGraph::Schema* s_;   // root schema file
    SemanticGraph::Schema* cur_; // current schema file


    //
    //
    Cult::Containers::Stack<SemanticGraph::Scope*> scope_stack_;


    //
    //
    struct Cardinality
    {
      Cardinality (UnsignedLong min, UnsignedLong max)
          : min_ (min), max_ (max)
      {
      }

      UnsignedLong min_, max_;
    };

    Cult::Containers::Stack<Cardinality> cardinality_stack_;

    // Map of absolute file path to a Schema node.
    //
    typedef
    Cult::Containers::Map<SemanticGraph::Path,
                          SemanticGraph::Schema*,
                          FilePathComparator>
    FileMap;

    FileMap file_map_;

    // Path stack for diagnostic.
    //
    Cult::Containers::Stack<SemanticGraph::Path> file_stack_;

    SemanticGraph::Path const&
    file ()
    {
      return file_stack_.top ();
    }

  private:
    Boolean qualify_attribute_;
    Boolean qualify_element_;

    Boolean valid_;
  };


  Parser::Impl::
  Impl ()
      : s_ (0),
        cur_ (0),
        qualify_attribute_ (false),
        qualify_element_ (false)
  {
  }

  template<typename T> T&
  add_type (Schema& s, Namespace& ns, String name)
  {
    Path file ("XMLSchema.xsd");
    T& node (s.new_node<T> (file, 0, 0));
    s.new_edge<Names> (ns, node, name);

    return node;
  }

  Evptr<Schema> Parser::Impl::
  parse (Path const& tu)
  {
    valid_ = true;

    XML::AutoPtr<Xerces::DOMDocument> d (dom (tu));

    if (!d)
      throw InvalidSchema ();

    XML::Element root (d->getDocumentElement ());
    String ns (root[L"targetNamespace"]);

    if (trace_)
      wcout << "target namespace: " << ns << endl;

    Evptr<Schema> rs (new Schema (tu, root.line (), root.column ()));

    // Implied schema with fundamental types.
    //
    {
      Path file ("XMLSchema.xsd");

      Schema& s (rs->new_node<Schema> (file, 1, 1));
      rs->new_edge<Implies> (*rs, s, file);

      Namespace& ns (s.new_node<Namespace> (file, 1, 1));
      s.new_edge<Names> (s, ns, xsd);

      AnyType& any_type (
        add_type<AnyType>                       (s, ns, L"anyType"));
      add_type<AnySimpleType>                   (s, ns, L"anySimpleType");


      // Integers.
      //
      add_type<Fundamental::Byte>               (s, ns, L"byte");
      add_type<Fundamental::UnsignedByte>       (s, ns, L"unsignedByte");
      add_type<Fundamental::Short>              (s, ns, L"short");
      add_type<Fundamental::UnsignedShort>      (s, ns, L"unsignedShort");
      add_type<Fundamental::Int>                (s, ns, L"int");
      add_type<Fundamental::UnsignedInt>        (s, ns, L"unsignedInt");
      add_type<Fundamental::Long>               (s, ns, L"long");
      add_type<Fundamental::UnsignedLong>       (s, ns, L"unsignedLong");
      add_type<Fundamental::Integer>            (s, ns, L"integer");
      add_type<Fundamental::NonPositiveInteger> (s, ns, L"nonPositiveInteger");
      add_type<Fundamental::NonNegativeInteger> (s, ns, L"nonNegativeInteger");
      add_type<Fundamental::PositiveInteger>    (s, ns, L"positiveInteger");
      add_type<Fundamental::NegativeInteger>    (s, ns, L"negativeInteger");


      // Boolean.
      //
      add_type<Fundamental::Boolean>            (s, ns, L"boolean");


      // Floats.
      //
      add_type<Fundamental::Float>              (s, ns, L"float");
      add_type<Fundamental::Double>             (s, ns, L"double");
      add_type<Fundamental::Decimal>            (s, ns, L"decimal");


      // Strings
      //
      add_type<Fundamental::String>             (s, ns, L"string");
      add_type<Fundamental::NormalizedString>   (s, ns, L"normalizedString");
      add_type<Fundamental::Token>              (s, ns, L"token");
      add_type<Fundamental::Name>               (s, ns, L"Name");
      add_type<Fundamental::NameToken>          (s, ns, L"NMTOKEN");
      add_type<Fundamental::NameTokens>         (s, ns, L"NMTOKENS");
      add_type<Fundamental::NCName>             (s, ns, L"NCName");
      add_type<Fundamental::Language>           (s, ns, L"language");

      // Qualified name.
      //
      add_type<Fundamental::QName>              (s, ns, L"QName");

      // ID/IDREF.
      //
      add_type<Fundamental::Id>                 (s, ns, L"ID");

      Fundamental::IdRef& id_ref (
        s.new_node<Fundamental::IdRef> (file, 0, 0));
      s.new_edge<Names> (ns, id_ref, L"IDREF");
      s.new_edge<Arguments> (any_type, id_ref);

      Fundamental::IdRefs& id_refs (
        s.new_node<Fundamental::IdRefs> (file, 0, 0));
      s.new_edge<Names> (ns, id_refs, L"IDREFS");
      s.new_edge<Arguments> (any_type, id_refs);


      // URI.
      //
      add_type<Fundamental::AnyURI>             (s, ns, L"anyURI");


      // Binary.
      //
      add_type<Fundamental::Base64Binary>       (s, ns, L"base64Binary");
      add_type<Fundamental::HexBinary>          (s, ns, L"hexBinary");


      // Date/time.
      //
      add_type<Fundamental::Date>               (s, ns, L"date");
      add_type<Fundamental::DateTime>           (s, ns, L"dateTime");
      add_type<Fundamental::Duration>           (s, ns, L"duration");
      add_type<Fundamental::Day>                (s, ns, L"gDay");
      add_type<Fundamental::Month>              (s, ns, L"gMonth");
      add_type<Fundamental::MonthDay>           (s, ns, L"gMonthDay");
      add_type<Fundamental::Year>               (s, ns, L"gYear");
      add_type<Fundamental::YearMonth>          (s, ns, L"gYearMonth");
      add_type<Fundamental::Time>               (s, ns, L"time");


      // Entity.
      //
      add_type<Fundamental::Entity>             (s, ns, L"ENTITY");
      add_type<Fundamental::Entities>           (s, ns, L"ENTITIES");


      // Notation.
      //
      add_type<Fundamental::Notation>           (s, ns, L"NOTATION");
    }


    // Parse.
    //
    {
      // Enter the file into file_map_.
      //
      file_map_[system_complete (tu)] = rs.get ();


      s_ = cur_ = rs.get ();
      {
        file_stack_.push (tu);

        {
          push_scope (
            s_->new_node<Namespace> (
              file (), root.line (), root.column ()));
          s_->new_edge<Names> (*cur_, scope (), ns);

          {
            schema (root);
          }

          pop_scope ();
        }

        file_stack_.pop ();
      }

      s_ = cur_ = 0;
    }

    // Second passes to resolve types, refs and groups.
    //
    for (UnsignedLong i (0); valid_; ++i)
    {
      Traversal::Schema schema;

      struct Contains : Traversal::Contains
      {
        virtual Void
        traverse (Type& e)
        {
          SemanticGraph::Node& n (e.element ());

          if (!n.context ().count ("seen"))
          {
            n.context ().set ("seen", true);
            Traversal::Contains::dispatch (n);
            n.context ().remove ("seen");
          }
        }
      } contains;

      Traversal::Names schema_names;
      Traversal::Namespace ns;

      schema >> contains;
      schema >> schema_names;
      schema_names >> ns;
      contains >> schema;

      Traversal::Names ns_names;
      Resolver resolver (*rs, valid_);

      // If I put it inside AnonymousType, gdb will puke.
      //
      struct Instance: Traversal::Instance
      {
        Instance (Traversal::NodeDispatcherBase& scope)
        {
          belongs_.node_traverser (scope);
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.typed () &&
              !e.type ().named () &&
              !e.type ().context ().count ("seen"))
          {
            e.type().context ().set ("seen", true);

            belongs (e, belongs_);

            e.type ().context ().remove ("seen");
          }
        }
      private:
        Traversal::Belongs belongs_;
      };

      struct AnonymousType : Traversal::Names
      {
        AnonymousType (Traversal::NodeDispatcherBase& scope)
            : instance_ (scope)
        {
          *this >> instance_;
        }

      private:


        Instance instance_;

      } anonymous_type (resolver);

      ns >> anonymous_type;
      ns >> ns_names >> resolver;

      Traversal::Names names;
      resolver >> anonymous_type;
      resolver >> names >> resolver;

      if (trace_)
        wcout << "post-resolution pass #" << i << endl;

      schema.dispatch (*rs);

      if (!resolver.need_more ())
        break;
    }

    if (!valid_)
      throw InvalidSchema ();

    return rs;
  }


  Void Parser::Impl::
  schema (XML::Element const& s)
  {
    Boolean old_qa (qualify_attribute_);
    Boolean old_qe (qualify_element_);

    if (String af = s[L"attributeFormDefault"])
      qualify_attribute_ = af == L"qualified";

    if (String ef = s[L"elementFormDefault"])
      qualify_element_ = ef == L"qualified";

    push (s);

    while (more ())
    {
      XML::Element e (next ());
      String name (e.name ());

      if (trace_)
        wcout << name << endl;

      if (name == L"group")          group (e);           else
      if (name == L"import")         import (e);          else
      if (name == L"include")        include (e);         else
      if (name == L"element")        element (e, true);   else
      if (name == L"attribute")      attribute (e, true); else
      if (name == L"simpleType")     simple_type (e);     else
      if (name == L"annotation");                         else
      if (name == L"complexType")    complex_type (e);    else
      if (name == L"attributeGroup") attribute_group (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: unexpected top-level element: '" << name << "'"
              << endl;

        valid_ = false;
      }
    }

    pop ();

    qualify_attribute_ = old_qa;
    qualify_element_ = old_qe;
  }

  Void Parser::Impl::
  import (XML::Element const& i)
  {
    String loc (i[L"schemaLocation"]);

    Path path, rel_path, abs_path;
    {
      //@@ Any better way?
      //
      Size len (std::wcstombs (0, loc.c_str (), 0) + 1);
      Char* str (new Char[len]);
      std::wcstombs (str, loc.c_str (), len);

      try
      {
        path = Path (str);
        rel_path = file ().branch_path () / path;
        abs_path = system_complete (rel_path);
        abs_path.normalize ();
      }
      catch (InvalidPath const&)
      {
        wcerr << file () << ":" << i.line () << ":" << i.column () << ": "
              << ": error: '" << str << "' is not a valid filesystem path"
              << endl;

        delete[] str;

        valid_ = false;
        return;
      }

      delete[] str;
    }

    if (file_map_.find (abs_path) != file_map_.end ())
    {
      s_->new_edge<Imports> (*cur_, *file_map_[abs_path], path);
      return;
    }

    if (trace_)
      wcout << "importing " << rel_path << endl;

    if (XML::AutoPtr<Xerces::DOMDocument> d  = dom (abs_path))
    {
      XML::Element r (d->getDocumentElement ());
      String ns (r[L"targetNamespace"]);

      if (trace_)
        wcout << "target namespace: " << ns << endl;

      Schema& s (s_->new_node<Schema> (rel_path, r.line (), r.column ()));
      s_->new_edge<Imports> (*cur_, s, path);

      file_map_[abs_path] = &s;

      Schema* old (cur_);
      cur_ = &s;

      {
        file_stack_.push (rel_path);

        {
          push_scope (
            s_->new_node<Namespace> (file (), r.line (), r.column ()));
          s_->new_edge<Names> (*cur_, scope (), ns);

          {
            schema (r);
          }

          pop_scope ();
        }

        file_stack_.pop ();
      }

      cur_ = old;
    }
  }

  Void Parser::Impl::
  include (XML::Element const& i)
  {
    String loc (i[L"schemaLocation"]);

    Path path, rel_path, abs_path;
    {
      //@@ Any better way?
      //
      Size len (std::wcstombs (0, loc.c_str (), 0) + 1);
      Char* str (new Char[len]);
      std::wcstombs (str, loc.c_str (), len);

      try
      {
        path = Path (str);
        rel_path = file ().branch_path () / path;
        abs_path = system_complete (rel_path);
        abs_path.normalize ();
      }
      catch (InvalidPath const&)
      {
        wcerr << file () << ":" << i.line () << ":" << i.column () << ": "
              << ": error: '" << str << "' is not a valid filesystem path"
              << endl;

        delete[] str;

        valid_ = false;
        return;
      }

      delete[] str;
    }

    if (file_map_.find (abs_path) != file_map_.end ())
    {
      Schema& s (*file_map_[abs_path]);

      // Here I am using the fact that each Schema Names only one
      // Namespace.
      //
      String ns (s.names_begin ()->name ()), cur_ns;

      if (ns.empty () && !(cur_ns = (cur_->names_begin ())->name ()).empty ())
      {
        // Chameleon.
        //
        // @@ I need to check if the acquired namespaces are the same.
        //
        s_->new_edge<Sources> (*cur_, s, path);
      }
      else
        s_->new_edge<Includes> (*cur_, s, path);

      return;
    }

    if (trace_)
      wcout << "including " << rel_path << endl;

    if (XML::AutoPtr<Xerces::DOMDocument> d  = dom (abs_path))
    {
      XML::Element r (d->getDocumentElement ());
      String ns (r[L"targetNamespace"]), cur_ns;

      Schema& s (s_->new_node<Schema> (rel_path, r.line (), r.column ()));

      file_map_[abs_path] = &s;

      if (ns.empty () && !(cur_ns = (cur_->names_begin ())->name ()).empty ())
      {
        // Chameleon.
        //
        ns = cur_ns;
        s_->new_edge<Sources> (*cur_, s, path);

        if (trace_)
          wcout << "handling chameleon schema" << endl;
      }
      else
        s_->new_edge<Includes> (*cur_, s, path);

      if (trace_)
        wcout << "target namespace: " << ns << endl;

      Schema* old (cur_);
      cur_ = &s;

      {
        file_stack_.push (rel_path);

        {
          push_scope (
            s_->new_node<Namespace> (file (), r.line (), r.column ()));
          s_->new_edge<Names> (*cur_, scope (), ns);

          {
            schema (r);
          }

          pop_scope ();
        }

        file_stack_.pop ();
      }

      cur_ = old;
    }
  }

  Void Parser::Impl::
  group (XML::Element const& g)
  {
    if (String name = g[L"name"])
    {
      Scope& group (s_->new_node<Scope> (file (), g.line (), g.column ()));
      s_->new_edge<Names> (scope (), group, name);

      push_scope (group);
      push (g);

      annotation ();

      XML::Element e (next ());

      name = e.name ();

      if (trace_)
        wcout << name << endl;

      if (name == L"all")      all (e);      else
      if (name == L"choice")   choice (e);   else
      if (name == L"sequence") sequence (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: expected 'all', 'choice' or 'sequence' "
              << "instead of '" << name << "'" << endl;

        valid_ = false;
      }

      pop ();
      pop_scope ();
    }
    else if (String ref = g[L"ref"])
    {
      //@@ min/max
      //
      // min/max can come from element in the group!

      UnsignedLong
        min (g[L"minOccurs"] == L"0" ? 0 : this->min ()),
        max (g[L"maxOccurs"] && g[L"maxOccurs"] != L"1" ? 0 : this->max ());

      if (trace_)
        wcout << "group min " << min << " max " << max << endl;

      try
      {
        String uq_name (XML::uq_name (ref));
        String ns_name;

        try
        {
          ns_name = XML::ns_name (g, ref);
        }
        catch (XML::NoMapping const& ex)
        {
          if (ex.prefix ().empty () && !has_target_ns ())
            ns_name = cur_->names_begin ()->name ();
          else
            throw;
        }

        Scope& s (resolve<Scope> (ns_name, uq_name, *s_));

        for (Scope::NamesIterator i (s.names_begin ());
             i != s.names_end (); ++i)
        {
          // Normally a referenced Group would be replaced with its
          // content. However it is not possible when the Group hasn't
          // been defined yet. In this case we simply copy the group
          // here and let resolver handle this.
          //
          if (dynamic_cast<Scope*> (&(i->named ())))
          {
            Scope& p (dynamic_cast<Scope&> (i->named ()));
            Scope& ref (s_->new_node<Scope> (file (), g.line (), g.column ()));

            String ns_name (p.context ().get<String> ("group-ns-name"));
            String uq_name (p.context ().get<String> ("group-uq-name"));
            UnsignedLong pmin (p.context ().get<UnsignedLong> ("group-min"));
            UnsignedLong pmax (p.context ().get<UnsignedLong> ("group-max"));

            ref.context ().set ("group-ns-name", ns_name);
            ref.context ().set ("group-uq-name", uq_name);
            ref.context ().set ("group-min", pmin == 0 ? 0 : min);
            ref.context ().set ("group-max", pmax != 1 ? pmax : max);

            s_->new_edge<Names> (scope (), ref, L"");
          }
          else
          {
            Element& prot (dynamic_cast<Element&> (i->named ()));

            Element& e (
              s_->new_node<Element> (prot.file (),
                                     prot.line (),
                                     prot.column (),
                                     min == 0 ? min : prot.min (),
                                     max == 0 ? max : prot.max (),
                                     prot.global (),
                                     prot.qualified ()));

            s_->new_edge<Names> (scope (), e, prot.name ());

            if (prot.qualified () &&
                !prot.context ().count ("instance-ns-name"))
              s_->new_edge<BelongsToNamespace> (e, prot.namespace_ ());

            // Now we set element's type.
            //
            if (prot.typed ())
            {
              // Easy case.
              //
              s_->new_edge<Belongs> (e, prot.type ());
            }
            else if (prot.context ().count ("type-ns-name"))
            {
              String ns_name (prot.context ().get<String> ("type-ns-name"));
              String uq_name (prot.context ().get<String> ("type-uq-name"));

              e.context ().set ("type-ns-name", ns_name);
              e.context ().set ("type-uq-name", uq_name);
            }
            else if (prot.context ().count ("instance-ns-name"))
            {
              String ns_name (prot.context ().get<String> ("instance-ns-name"));
              String uq_name (prot.context ().get<String> ("instance-uq-name"));

              e.context ().set ("instance-ns-name", ns_name);
              e.context ().set ("instance-uq-name", uq_name);
            }
            else if (valid_)
            {
              wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
                    << "ice: element is in unexpected condition" << endl;

              abort ();
            }
          }
        }
      }
      catch (NotNamespace const& ex)
      {
        if (valid_)
        {
          wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
                << "ice: unable to resolve namespace '" << ex.ns () << "'"
                << endl;

          abort ();
        }
      }
      catch (NotName const& ex)
      {
        if (trace_)
          wcout << "unable to resolve group name '" << ex.name ()
                << "' inside namespace '" << ex.ns () << "'" << endl
                << "deferring resolution until later" << endl;

        // Cannot resolve group at this point. Will have to try later.
        // For now insert a dummy Scope with all the information stored
        // in its context.
        //
        Scope& ref (s_->new_node<Scope> (file (), g.line (), g.column ()));

        ref.context ().set ("group-ns-name", ex.ns ());
        ref.context ().set ("group-uq-name", ex.name ());
        ref.context ().set ("group-min", min);
        ref.context ().set ("group-max", max);

        s_->new_edge<Names> (scope (), ref, L"");
      }
      catch (XML::NoMapping const& ex)
      {
        wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
              << "unable to resolve namespace prefix '" << ex.prefix () << "' "
              << "in '" << ref << "'" << endl;

        valid_ = false;
      }
    }
    else
    {
      wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
            << "error: 'name' or 'ref' attribute is missing in group "
            << "declaration" << endl;

      valid_ = false;

      return;
    }
  }

  //@@ Need RAII for push/pop.
  //

  Type* Parser::Impl::
  simple_type (XML::Element const& t)
  {
    Type* r (0);

    push (t);

    annotation ();

    XML::Element e (next ());

    String name (e.name ());

    if (name == L"list") r = list (e, t); else
    if (name == L"restriction") r = restriction (e, t); else
    {
      wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
            << "error: expected 'list' or 'restriction' "
            << "instead of '" << name << "'" << endl;

      valid_ = false;
    }

    pop ();

    return r;
  }

  SemanticGraph::Type* Parser::Impl::
  list (XML::Element const& l, XML::Element const& t)
  {
    String item_type (l[L"itemType"]);

    if (!item_type)
    {
      wcerr << file () << ":" << l.line () << ":" << l.column () << ": "
            << "error: expected 'itemType' attribute" << endl;

      wcerr << file () << ":" << l.line () << ":" << l.column () << ": "
            << "info: embedded simpleType definition is not supported"
            << endl;

      valid_ = false;

      return 0;
    }

    if (trace_)
      wcout << "list item type: " << fq_name (l, item_type) << endl;

    List& node (s_->new_node<List> (file (), t.line (), t.column ()));
    set_type<Arguments> (item_type, l, node);

    if (String name = t[L"name"])
      s_->new_edge<Names> (scope (), node, name);

    return &node;
  }

  Type* Parser::Impl::
  restriction (XML::Element const& r, XML::Element const& t)
  {
    String base (r[L"base"]);

    if (!base)
    {
      wcerr << file () << ":" << r.line () << ":" << r.column () << ": "
            << "error: expected 'base' attribute" << endl;

      wcerr << file () << ":" << r.line () << ":" << r.column () << ": "
            << "info: embedded simpleType definition is not supported"
            << endl;

      valid_ = false;

      return 0;
    }

    Type* rv (0);

    if (trace_)
      wcout << "restriction base: " << fq_name (r, base) << endl;

    push (r);

    annotation ();

    Boolean enum_ (false);

    if (more ())
    {
      XML::Element e (next ());

      if (e.name () == L"enumeration")
      {
        // Enumeration
        //
        enum_ = true;

        Enumeration& node (
          s_->new_node<Enumeration> (file (), t.line (), t.column ()));

        set_type<Inherits> (base, r, node);

        if (String name = t[L"name"])
          s_->new_edge<Names> (scope (), static_cast<Nameable&> (node), name);

        push_scope (node);

        for (enumeration (e); more (); enumeration (next ()));

        pop_scope ();

        rv = &node;
      }
    }

    if (!enum_)
    {
      Complex& node (s_->new_node<Complex> (file (), t.line (), t.column ()));
      set_type<Inherits> (base, r, node);

      if (String name = t[L"name"])
        s_->new_edge<Names> (scope (), node, name);

      rv = &node;
    }

    pop ();

    return rv;
  }

  Void Parser::Impl::
  enumeration (XML::Element const& e)
  {
    String value (e[L"value"]);

    if (trace_)
      wcout << "enumeration value: " << value << endl;

    Enumerator& node (
      s_->new_node<Enumerator> (file (), e.line (), e.column ()));

    s_->new_edge<Names> (scope (), node, value);
    s_->new_edge<Belongs> (node, dynamic_cast<Type&>(scope ()));
  }

  Type* Parser::Impl::
  complex_type (XML::Element const& t)
  {
    Type* r (0);

    if (t[L"mixed"] == L"true")
    {
      /*
      wcerr << file () << ":" << t.line () << ":" << t.column () << ": "
            << "error: mixed content model is not supported" << endl;

      valid_ = false;
      return r;
      */
    }

    Complex& node (s_->new_node<Complex> (file (), t.line (), t.column ()));

    if (String name = t[L"name"])
      s_->new_edge<Names> (scope (), node, name);

    r = &node;

    push_scope (node);
    push (t);

    annotation ();

    if (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (trace_)
        wcout << name << endl;

      if (name == L"simpleContent")  simple_content (e);  else
      if (name == L"complexContent") complex_content (e); else
      {
        if (name == L"all")            all (e);             else
        if (name == L"group")          group (e);           else
        if (name == L"choice")         choice (e);          else
        if (name == L"sequence")       sequence (e);        else
        if (name == L"attribute")      attribute (e);       else
        if (name == L"anyAttribute")   any_attribute (e);   else
        if (name == L"attributeGroup") attribute_group (e); else
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "error: unexpected element '" << name << "'" << endl;

          valid_ = false;
        }

        while (more ())
        {
          XML::Element e (next ());
          String name (e.name ());

          if (name == L"attribute")      attribute (e);       else
          if (name == L"anyAttribute")   any_attribute (e);   else
          if (name == L"attributeGroup") attribute_group (e); else
          {
            wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                  << "error: expected 'attribute', 'anyAttribute', or "
                  << "'attributeGroup' instead of '" << name << "'" << endl;

            valid_ = false;
          }
        }
      }
    }

    pop ();
    pop_scope ();

    return r;
  }

  Void Parser::Impl::
  all (XML::Element const& a)
  {
    // 'all' cannot be nested inside 'choice' or 'sequence', nor
    //  can it contain any of those. The default cardinality
    //  values for 'all' are min=1, max=1. Therefore we don't
    //  really need to push/pop cardinality.
    //
    push (a);

    annotation ();

    while (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (name == L"element") element (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: expected 'element' "
              << "instead of '" << name << "'" << endl;

        valid_ = false;
      }
    }

    pop ();
  }

  Void Parser::Impl::
  choice (XML::Element const& c)
  {
    // Establish cardinality.
    //
    push_cardinality (
      0, c[L"maxOccurs"] && c[L"maxOccurs"] != L"1" ? 0 : max ());

    push (c);

    annotation ();

    while (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (name == L"any")      any (e);      else
      if (name == L"group")    group (e);    else
      if (name == L"choice")   choice (e);   else
      if (name == L"element")  element (e);  else
      if (name == L"sequence") sequence (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: expected 'any', 'group', 'choice', 'sequence', "
              << "or 'element' instead of '" << name << "'" << endl;

        valid_ = false;
      }
    }

    pop ();

    pop_cardinality ();
  }

  Void Parser::Impl::
  sequence (XML::Element const& s)
  {
    // Establish cardinality.
    //
    push_cardinality (
      s[L"minOccurs"] == L"0" ? 0 : min (),
      s[L"maxOccurs"] && s[L"maxOccurs"] != L"1" ? 0 : max ());

    push (s);

    annotation ();

    while (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (name == L"any")      any (e);      else
      if (name == L"group")    group (e);    else
      if (name == L"choice")   choice (e);   else
      if (name == L"element")  element (e);  else
      if (name == L"sequence") sequence (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: expected 'any', 'group', 'choice', 'sequence', "
              << "or 'element' instead of '" << name << "'" << endl;

        valid_ = false;
      }
    }

    pop ();

    pop_cardinality ();
  }

  Void Parser::Impl::
  simple_content (XML::Element const& c)
  {
    push (c);

    XML::Element e (next ());
    String name (e.name ());

    if (name == L"extension")   simple_content_extension (e);   else
    if (name == L"restriction") simple_content_restriction (e); else
    {
      wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
            << "error: expected 'extension' or 'restriction' instead of "
            << "'" << name << "'" << endl;

      valid_ = false;
    }

    pop ();
  }

  Void Parser::Impl::
  complex_content (XML::Element const& c)
  {
    if (c[L"mixed"] == L"true")
    {
      /*
      wcerr << file () << ":" << c.line () << ":" << c.column () << ": "
            << "error: mixed content model is not supported" << endl;


      valid_ = false;
      */
    }

    push (c);

    XML::Element e (next ());
    String name (e.name ());

    if (name == L"extension")   complex_content_extension (e);   else
    if (name == L"restriction") complex_content_restriction (e); else
    {
      wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
            << "error: expected 'extension' or 'restriction' instead of "
            << "'" << name << "'" << endl;

      valid_ = false;
    }

    pop ();
  }

  Void Parser::Impl::
  simple_content_extension (XML::Element const& e)
  {
    if (trace_)
      wcout << "extension base: " << fq_name (e, e[L"base"]) << endl;

    set_type<Inherits> (e[L"base"], e, dynamic_cast<Complex&> (scope ()));

    push (e);

    while (more ())
    {
      XML::Element e (next ());
      String name (e.name ());

      if (name == L"attribute")      attribute (e);       else
      if (name == L"anyAttribute")   any_attribute (e);   else
      if (name == L"attributeGroup") attribute_group (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: expected 'attribute', 'anyAttribute', or "
              << "'attributeGroup' instead of '" << name << "'" << endl;

        valid_ = false;
      }
    }

    pop ();
  }

  Void Parser::Impl::
  simple_content_restriction (XML::Element const& e)
  {
    if (trace_)
      wcout << "restriction base: " << fq_name (e, e[L"base"]) << endl;

    set_type<Inherits> (e[L"base"], e, dynamic_cast<Complex&> (scope ()));

    // We simply ignore the contents for now.
    //
  }

  Void Parser::Impl::
  complex_content_extension (XML::Element const& e)
  {
    if (trace_)
      wcout << "extension base: " << fq_name (e, e[L"base"]) << endl;

    set_type<Inherits> (e[L"base"], e, dynamic_cast<Complex&> (scope ()));

    push (e);

    if (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (name == L"all")            all (e);             else
      if (name == L"group")          group (e);           else
      if (name == L"choice")         choice (e);          else
      if (name == L"sequence")       sequence (e);        else
      if (name == L"attribute")      attribute (e);       else
      if (name == L"anyAttribute")   any_attribute (e);   else
      if (name == L"attributeGroup") attribute_group (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: unexpected element '" << name << "'" << endl;

        valid_ = false;
      }

      while (more ())
      {
        XML::Element e (next ());
        String name (e.name ());

        if (name == L"attribute")      attribute (e);       else
        if (name == L"anyAttribute")   any_attribute (e);   else
        if (name == L"attributeGroup") attribute_group (e); else
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "error: expected 'attribute', 'anyAttribute', or "
                << "'attributeGroup' instead of '" << name << "'" << endl;

          valid_ = false;
        }
      }
    }

    pop ();
  }

  Void Parser::Impl::
  complex_content_restriction (XML::Element const& e)
  {
    if (trace_)
      wcout << "restriction base: " << fq_name (e, e[L"base"]) << endl;

    set_type<Inherits> (e[L"base"], e, dynamic_cast<Complex&> (scope ()));

    // @@
    // For now we simply skip the contents unless the base is anyType
    // (or a trivial alias thereof). Checking for the trivial alias
    // is further complicated by the fact that it might not be defined
    // at this stage (forward inheritnace) so we will ignore that case
    // as well for now.
    //
    {
      String base (e[L"base"]);
      String uq_name (XML::uq_name (base));
      String ns_name;

      try
      {
        ns_name = XML::ns_name (e, base);
      }
      catch (XML::NoMapping const& ex)
      {
        if (ex.prefix ().empty () && !has_target_ns ())
          ns_name = cur_->names_begin ()->name ();
        else
          return;
      }

      if (ns_name != xsd || uq_name != L"anyType")
        return;
    }

    push (e);

    if (more ())
    {
      XML::Element e (next ());

      String name (e.name ());

      if (name == L"all")            all (e);             else
      if (name == L"group")          group (e);           else
      if (name == L"choice")         choice (e);          else
      if (name == L"sequence")       sequence (e);        else
      if (name == L"attribute")      attribute (e);       else
      if (name == L"anyAttribute")   any_attribute (e);   else
      if (name == L"attributeGroup") attribute_group (e); else
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: unexpected element '" << name << "'" << endl;

        valid_ = false;
      }

      while (more ())
      {
        XML::Element e (next ());
        String name (e.name ());

        if (name == L"attribute")      attribute (e);       else
        if (name == L"anyAttribute")   any_attribute (e);   else
        if (name == L"attributeGroup") attribute_group (e); else
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "error: expected 'attribute', 'anyAttribute', or "
                << "'attributeGroup' instead of '" << name << "'" << endl;

          valid_ = false;
        }
      }
    }

    pop ();
  }

  Void Parser::Impl::
  element (XML::Element const& e, Boolean global)
  {
    UnsignedLong min (
      e[L"minOccurs"] == L"0" ? 0 : this->min ()),
      max (e[L"maxOccurs"] && e[L"maxOccurs"] != L"1" ? 0 : this->max ());

    Boolean qualified (global ? true : qualify_element_);

    if (String form = e[L"form"])
      qualified = form == L"qualified";

    if (trace_)
      wcout << "element min " << min << " max " << max << " qualified "
            << qualified << endl;

    if (String name = e[L"name"])
    {
      if (trace_)
        wcout << "element name '" << name << "'" << endl;

      Element& node (
        s_->new_node<Element> (
          file (), e.line (), e.column (), min, max, global, qualified));

      s_->new_edge<Names> (scope (), node, name);

      if (qualified)
        s_->new_edge<BelongsToNamespace> (node, cur_ns ());

      if (global)
      {
        if (String sg = e[L"substitutionGroup"])
        {
          if (trace_)
            wcout << "substitutes " << sg << endl;

          String uq_name (XML::uq_name (sg));
          String ns_name;

          try
          {
            try
            {
              ns_name = XML::ns_name (e, sg);
            }
            catch (XML::NoMapping const& ex)
            {
              if (ex.prefix ().empty () && !has_target_ns ())
                ns_name = cur_->names_begin ()->name ();
              else
                throw;
            }

            node.context ().set ("substitution-ns-name", ns_name);
            node.context ().set ("substitution-uq-name", uq_name);
          }
          catch (XML::NoMapping const& ex)
          {

            wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                  << "unable to resolve namespace prefix '" << ex.prefix () << "' "
                  << "in '" << sg << "'" << endl;

            valid_ = false;
          }
        }
      }

      if (String type = e[L"type"])
      {
        if (trace_)
          wcout << "element type " << fq_name (e, type) << endl;

        set_type<Belongs> (type, e, node);
      }
      else
      {
        // Looks like an anonymous type.
        //
        push (e);

        annotation ();

        if (more ())
        {
          XML::Element e (next ());

          String name (e.name ());

          if (trace_)
            wcout << name << endl;

          Type* t (0);

          if (name == L"simpleType")  t = simple_type (e);  else
          if (name == L"complexType") t = complex_type (e); else
          {
            wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                  << "error: expected 'simpleType' or 'complexType' "
                  << "instead of '" << e.name () << "'" << endl;

            valid_ = false;
          }

          if (t)
            s_->new_edge<Belongs> (node, *t);
        }
        else
        {
          // anyType
          //

          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "warning: element '" << name << "' is implicitly "
                << " of anyType" << endl;

          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "info: did you forget to specify 'type' attribute?"
                << endl;

          String prefix (ns_prefix (e, xsd));
          type =  prefix + (prefix.empty () ? L"" : L":") + L"anyType";

          set_type<Belongs> (type, e, node);
        }

        pop ();
      }
    }
    else if (String ref = e[L"ref"])
    {
      Element& node (
        s_->new_node<Element> (
          file (), e.line (), e.column (), min, max, true, true));
      try
      {
        String uq_name (XML::uq_name (ref));
        s_->new_edge<Names> (scope (), node, uq_name);

        String ns_name;

        try
        {
          ns_name = XML::ns_name (e, ref);
        }
        catch (XML::NoMapping const& ex)
        {
          if (ex.prefix ().empty () && !has_target_ns ())
            ns_name = cur_->names_begin ()->name ();
          else
            throw;
        }

        Element& prototype (resolve<Element> (ns_name, uq_name, *s_));

        s_->new_edge<BelongsToNamespace> (node, prototype.namespace_ ());

        if (prototype.typed ())
        {
          s_->new_edge<Belongs> (node, prototype.type ());
        }
        else if (prototype.context ().count ("type-ns-name"))
        {
          String ns_name (prototype.context ().get<String> ("type-ns-name"));
          String uq_name (prototype.context ().get<String> ("type-uq-name"));

          node.context ().set ("type-ns-name", ns_name);
          node.context ().set ("type-uq-name", uq_name);

          if (trace_)
            wcout << "element '" << ref << "' is not typed" << endl
                  << "deferring resolution until later" << endl;
        }
        else if (valid_)
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "ice: element '" << ref << "' is in unexpected condition"
                << endl;

          abort ();
        }
      }
      catch (NotNamespace const& ex)
      {
        if (valid_)
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "ice: unable to resolve namespace '" << ex.ns () << "'"
                << endl;

          abort ();
        }
      }
      catch (NotName const& ex)
      {
        node.context ().set ("instance-ns-name", ex.ns ());
        node.context ().set ("instance-uq-name", ex.name ());

        if (trace_)
          wcout << "unable to resolve name '" << ex.name ()
                << "' inside namespace '" << ex.ns () << "'" << endl
                << "deferring resolution until later" << endl;
      }
      catch (XML::NoMapping const& ex)
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "unable to resolve namespace prefix '" << ex.prefix () << "' "
              << "in '" << ref << "'" << endl;

        valid_ = false;
      }
    }
    else
    {
      if (valid_)
      {
        wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
              << "error: 'name' or 'ref' attribute is missing in element "
              << "declaration" << endl;
      }
    }
  }

  Void Parser::Impl::
  annotation ()
  {
    if (more ())
    {
      if (next ().name () != L"annotation")
        prev ();
    }
  }


  Void Parser::Impl::
  attribute (XML::Element const& a, Boolean global)
  {
    Boolean optional (true);

    String use (a[L"use"]);

    if (use == L"prohibited")
      return;
    else if (use == L"required")
      optional = false;

    Boolean qualified (global ? true : qualify_attribute_);

    if (String form = a[L"form"])
      qualified = form == L"qualified";

    if (String name = a[L"name"])
    {
      if (trace_)
        wcout << "attribute '" << name << "'" << endl;

      Attribute& node (
        s_->new_node<Attribute> (
          file (), a.line (), a.column (), optional, global, qualified));

      s_->new_edge<Names> (scope (), node, name);

      if (qualified)
        s_->new_edge<BelongsToNamespace> (node, cur_ns ());

      if (String type = a[L"type"])
      {
        if (trace_)
          wcout << "attribute type: '" << fq_name (a, type) << "'" << endl;

        set_type<Belongs> (type, a, node);
      }
      else
      {
        // Looks like an anonymous type.
        //
        push (a);

        annotation ();

        if (more ())
        {
          XML::Element e (next ());

          String name (e.name ());

          if (trace_)
            wcout << name << endl;

          Type* t (0);

          if (name == L"simpleType") t = simple_type (e); else
          {
            wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
                  << "error: expected 'simpleType' instead of '" << e.name ()
                  << "'" << endl;

            valid_ = false;
          }

          if (t)
            s_->new_edge<Belongs> (node, *t);
        }
        else
        {
          wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
                << "warning: attribute '" << name << "' is implicitly "
                << " of anySimpleType" << endl;

          wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
                << "info: did you forget to specify 'type' attribute?"
                << endl;

          // anySimpleType
          //
          String prefix (ns_prefix (a, xsd));
          type =  prefix + (prefix.empty () ? L"" : L":") + L"anySimpleType";

          set_type<Belongs> (type, a, node);
        }

        pop ();
      }
    }
    else if (String ref = a[L"ref"])
    {
      Attribute& node (
        s_->new_node<Attribute> (
          file (), a.line (), a.column (), optional, true, true));

      try
      {
        String uq_name (XML::uq_name (ref));
        s_->new_edge<Names> (scope (), node, uq_name);

        String ns_name;

        try
        {
          ns_name = XML::ns_name (a, ref);
        }
        catch (XML::NoMapping const& ex)
        {
          if (ex.prefix ().empty () && !has_target_ns ())
            ns_name = cur_->names_begin ()->name ();
          else
            throw;
        }

        Attribute& prototype (resolve<Attribute> (ns_name, uq_name, *s_));

        s_->new_edge<BelongsToNamespace> (node, prototype.namespace_ ());

        if (prototype.typed ())
        {
          s_->new_edge<Belongs> (node, prototype.type ());
        }
        else if (prototype.context ().count ("type-ns-name"))
        {
          String ns_name (prototype.context ().get<String> ("type-ns-name"));
          String uq_name (prototype.context ().get<String> ("type-uq-name"));

          node.context ().set ("type-ns-name", ns_name);
          node.context ().set ("type-uq-name", uq_name);

          if (trace_)
            wcout << "attribute '" << ref << "' is not typed" << endl
                  << "deferring resolution until later" << endl;
        }
        else if (valid_)
        {
          wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
                << "ice: attribute '" << ref << "' is in unexpected condition"
                << endl;

          abort ();
        }
      }
      catch (NotNamespace const& ex)
      {
        if (valid_)
        {
          wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
                << "ice: unable to resolve namespace '" << ex.ns () << "'"
                << endl;

          abort ();
        }
      }
      catch (NotName const& ex)
      {
        node.context ().set ("instance-ns-name", ex.ns ());
        node.context ().set ("instance-uq-name", ex.name ());

        if (trace_)
          wcout << "unable to resolve name '" << ex.name ()
                << "' inside namespace '" << ex.ns () << "'" << endl
                << "deferring resolution until later" << endl;
      }
      catch (XML::NoMapping const& ex)
      {
        wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
              << "unable to resolve namespace prefix '" << ex.prefix () << "' "
              << "in '" << ref << "'" << endl;

        valid_ = false;
      }
    }
    else
    {
      if (valid_)
      {
        wcerr << file () << ":" << a.line () << ":" << a.column () << ": "
              << "error: 'name' or 'ref' attribute is missing in attribute "
              << "declaration" << endl;
      }
    }
  }

  Void Parser::Impl::
  attribute_group (XML::Element const& g)
  {
    if (String name = g[L"name"])
    {
      // Global definition.
      //
      if (trace_)
        wcout << "attributeGroup '" << name << "'" << endl;

      Scope& group (s_->new_node<Scope> (file (), g.line (), g.column ()));
      s_->new_edge<Names> (scope (), group, name);

      push_scope (group);
      push (g);

      annotation ();

      while (more ())
      {
        XML::Element e (next ());
        String name (e.name ());

        if (trace_)
          wcout << name << endl;

        if (name == L"attribute")      attribute (e);       else
        if (name == L"anyAttribute")   any_attribute (e);   else
        if (name == L"attributeGroup") attribute_group (e); else
        {
          wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
                << "error: expected 'attribute', 'anyAttribute', or "
                << "'attributeGroup' instead of '" << name << "'" << endl;

          valid_ = false;
        }
      }

      pop ();
      pop_scope ();
    }
    else if (String ref = g[L"ref"])
    {
      if (trace_)
        wcout << "attributeGroup ref='" << ref << "'" << endl;

      try
      {
        String uq_name (XML::uq_name (ref));
        String ns_name;

        try
        {
          ns_name = XML::ns_name (g, ref);
        }
        catch (XML::NoMapping const& ex)
        {
          if (ex.prefix ().empty () && !has_target_ns ())
            ns_name = cur_->names_begin ()->name ();
          else
            throw;
        }

        Scope& s (resolve<Scope> (ns_name, uq_name, *s_));

        for (Scope::NamesIterator i (s.names_begin ());
             i != s.names_end (); ++i)
        {
          // Normally a referenced attributeGroup would be replaced
          // with its content. However it is not possible when
          // attributeGroup hasn't been defined yet. In this case
          // we simply copy the group here and let resolver handle
          // this.
          //
          if (dynamic_cast<Scope*> (&(i->named ())))
          {
            Scope& p (dynamic_cast<Scope&> (i->named ()));
            Scope& ref (s_->new_node<Scope> (file (), g.line (), g.column ()));

            String ns_name (p.context ().get<String> ("attrgroup-ns-name"));
            String uq_name (p.context ().get<String> ("attrgroup-uq-name"));

            ref.context ().set ("attrgroup-ns-name", ns_name);
            ref.context ().set ("attrgroup-uq-name", uq_name);

            s_->new_edge<Names> (scope (), ref, L"");
          }
          else
          {
            // Prototype.
            //
            Attribute& p (dynamic_cast<Attribute&> (i->named ()));

            Attribute& a (
              s_->new_node<Attribute> (p.file (),
                                       p.line (),
                                       p.column (),
                                       p.optional (),
                                       p.global (),
                                       p.qualified ()));

            s_->new_edge<Names> (scope (), a, p.name ());

            if (p.qualified () &&
                !p.context ().count ("instance-ns-name"))
              s_->new_edge<BelongsToNamespace> (a, p.namespace_ ());

            // Set attribute type.
            //
            if (p.typed ())
            {
              // Easy case.
              //
              s_->new_edge<Belongs> (a, p.type ());
            }
            else if (p.context ().count ("type-ns-name"))
            {
              String ns_name (p.context ().get<String> ("type-ns-name"));
              String uq_name (p.context ().get<String> ("type-uq-name"));

              a.context ().set ("type-ns-name", ns_name);
              a.context ().set ("type-uq-name", uq_name);
            }
            else if (p.context ().count ("instance-ns-name"))
            {
              String ns_name (p.context ().get<String> ("instance-ns-name"));
              String uq_name (p.context ().get<String> ("instance-uq-name"));

              a.context ().set ("instance-ns-name", ns_name);
              a.context ().set ("instance-uq-name", uq_name);
            }
            else if (valid_)
            {
              wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
                    << "ice: attribute is in unexpected condition" << endl;

              abort ();
            }
          }
        }
      }
      catch (NotNamespace const& ex)
      {
        if (valid_)
        {
          wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
                << "ice: unable to resolve namespace '" << ex.ns () << "'"
                << endl;

          abort ();
        }
      }
      catch (NotName const& ex)
      {
        if (trace_)
          wcout << "unable to resolve attribute group name '" << ex.name ()
                << "' inside namespace '" << ex.ns () << "'" << endl
                << "deferring resolution until later" << endl;

        // Cannot resolve group at this point. Will have to try later.
        // For now insert a dummy Scope with all the information stored
        // in its context.
        //
        Scope& ref (s_->new_node<Scope> (file (), g.line (), g.column ()));

        ref.context ().set ("attrgroup-ns-name", ex.ns ());
        ref.context ().set ("attrgroup-uq-name", ex.name ());

        s_->new_edge<Names> (scope (), ref, L"");
      }
      catch (XML::NoMapping const& ex)
      {
        wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
              << "unable to resolve namespace prefix '" << ex.prefix () << "' "
              << "in '" << ref << "'" << endl;

        valid_ = false;
      }
    }
    else
    {
      wcerr << file () << ":" << g.line () << ":" << g.column () << ": "
            << "error: 'name' or 'ref' attribute is missing in "
            << "attributeGroup declaration" << endl;

      valid_ = false;

      return;
    }
  }

  Void Parser::Impl::
  any (XML::Element const& a)
  {
    if (trace_)
      wcout << "any" << endl;

    UnsignedLong min (
      a[L"minOccurs"] == L"0" ? 0 : this->min ()),
      max (a[L"maxOccurs"] && a[L"maxOccurs"] != L"1" ? 0 : this->max ());

    String namespaces (a[L"namespace"] ? a[L"namespace"] : L"##any");


    Any& any (s_->new_node<Any> (file (),
                                 a.line (),
                                 a.column (),
                                 min,
                                 max,
                                 namespaces));

    s_->new_edge<Names> (scope (), any, L""); // No name.
  }

  Void Parser::Impl::
  any_attribute (XML::Element const&)
  {
    if (trace_)
      wcout << "anyAttribute" << endl;

    // For now we are simply ignoring.
    //
  }

  // Some specializations to get edge orientations right.
  //

  template <typename Edge, typename Node>
  struct Orientation
  {
    static Void
    set_edge (Schema& s, Node& node, Type& type)
    {
      // By default it is node->edge
      //
      s.template new_edge<Edge> (node, type);
    }
  };

  template <typename Node>
  struct Orientation<Arguments, Node>
  {
    static Void
    set_edge (Schema& s, Node& node, Type& type)
    {
      // For Arguments it is type->node.
      //
      s.template new_edge<Arguments> (type, node);
    }
  };

  template <typename Edge, typename Node>
  Void Parser::Impl::
  set_type (String const& type, XML::Element const& e, Node& node)
  {
    try
    {
      String uq_name (XML::uq_name (type));
      String ns_name;

      try
      {
        ns_name = XML::ns_name (e, type);
      }
      catch (XML::NoMapping const& ex)
      {
        if (ex.prefix ().empty () && !has_target_ns ())
          ns_name = cur_->names_begin ()->name ();
        else
          throw;
      }

      Type& t (resolve<Type> (ns_name, uq_name, *s_));

      // See if it is an IDREF specialization.
      //
      if (ns_name == xsd && (uq_name == L"IDREF" || uq_name == L"IDREFS"))
      {
        // See if we've got 'xse:refType' attribute.
        //
        if (String ref_type = e.attribute (xse, "refType"))
        {
          if (trace_)
            wcout << "found refType attribute '" << ref_type << "'" << endl;

          //@@ It is a bit wasteful to create a new spcialization for
          //   each refType. Instead we could lookup the target type
          //   and then navigate through Arguments edges to see if this
          //   type already arguments specialization that we are intersted
          //   in. But for now I will simply the logic by creating a new
          //   specialization every time.
          //

          Specialization* spec (0);

          if (uq_name == L"IDREF")
            spec = &s_->new_node<Fundamental::IdRef> (
              file (), e.line (), e.column ());
          else
            spec = &s_->new_node<Fundamental::IdRefs> (
              file (), e.line (), e.column ());

          Orientation<Edge, Node>::set_edge (*s_, node, *spec);

          set_type<Arguments> (ref_type, e, *spec);
        }
        else
          Orientation<Edge, Node>::set_edge (*s_, node, t);
      }
      else
        Orientation<Edge, Node>::set_edge (*s_, node, t);
    }
    catch (NotNamespace const& ex)
    {
      wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
            << "error: unable to resolve namespace '" << ex.ns () << "'"
            << endl;

      valid_ = false;

    }
    catch (NotName const& ex)
    {
      node.context ().set ("type-ns-name", ex.ns ());
      node.context ().set ("type-uq-name", ex.name ());

      if (trace_)
        wcout << "unable to resolve name '" << ex.name ()
              << "' inside namespace '" << ex.ns () << "'" << endl
              << "deferring resolution until later" << endl;
    }
    catch (XML::NoMapping const& ex)
    {
      wcerr << file () << ":" << e.line () << ":" << e.column () << ": "
            << "error: unable to resolve namespace prefix "
            << "'" << ex.prefix () << "' in '" << type << "'" << endl;

      valid_ = false;
    }
  }

  // Xerces has a provision to associate a public id with input streams
  // that can later be used in diagnostics. Unfortunately, it doesn't
  // work. So we will have to keep our own track.
  //
  struct Context: public NonCopyable
  {
    // File map for diagnostic.
    //
    Path const&
    file (Path const& abs) const
    {
      FileMap::ConstIterator i (file_map_.find (abs));

      if (i != file_map_.end ())
      {
        return i->second;
      }
      else
      {
        return abs;
      }
    }

    Void
    map_file (Path const& abs, Path const& rel)
    {
      file_map_[abs] = rel;
    }

  private:
    typedef Cult::Containers::Map<Path, Path, FilePathComparator> FileMap;
    FileMap file_map_;
  };

  //
  //
  class ErrorHandler : public  Xerces::DOMErrorHandler
  {
  public:
    ErrorHandler (Boolean& valid, Context const& ctx)
        : valid_ (valid),
          ctx_ (ctx)
    {
    }

    virtual Boolean
    handleError (Xerces::DOMError const& e)
    {
      // Xerces likes to say "Fatal error encountered during schema scan".
      // We don't need this junk.
      //
      if (!valid_
          && e.getLocation ()->getLineNumber () == 0
          && e.getLocation ()->getColumnNumber () == 0)
        return true;


      XSDFrontend::SemanticGraph::Path abs_path (
        XML::transcode_to_narrow (e.getLocation ()->getURI ()),
        boost::filesystem::native);

      XSDFrontend::SemanticGraph::Path rel_path (ctx_.file (abs_path));

      wcerr << rel_path << ':'
            << e.getLocation ()->getLineNumber () << ':'
            << e.getLocation ()->getColumnNumber () << ": ";

      switch (e.getSeverity ())
      {
      case Xerces::DOMError::DOM_SEVERITY_WARNING:
        {
          wcerr << "warning: ";
          break;
        }
      default:
        {
          wcerr << "error: ";
          valid_ = false;
          break;
        }
      }

      wcerr << e.getMessage () << endl;

      return true;
    }

  private:
    Boolean& valid_;
    Context const& ctx_;
  };

  /*
  class ErrorHandler : public  Xerces::ErrorHandler
  {
  public:
    ErrorHandler (Boolean& valid)
        : valid_ (valid)
    {
    }

    struct Severity
    {
      enum Value
      {
        warning,
        error,
        fatal
      };
    };

    virtual void
    resetErrors()
    {
    }

  private:
    virtual void
    warning (Xerces::SAXParseException const& e)
    {
      handle (e, Severity::warning);
    }

    virtual void
    error (Xerces::SAXParseException const& e)
    {
      handle (e, Severity::error);
    }


    virtual void
    fatalError (Xerces::SAXParseException const& e)
    {
      handle (e, Severity::fatal);
    }

    virtual void
    handle (Xerces::SAXParseException const& e, Severity::Value severity)
    {
      // Xerces likes to say "Fatal error encountered during schema scan".
      // We don't need this junk.
      //
      if (!valid_
          && e.getLineNumber () == 0
          && e.getColumnNumber () == 0)
        return;

      XMLCh const* id (e.getPublicId () ? e.getPublicId () : e.getSystemId ());

      wcerr << id << ':'
            << e.getLineNumber () << ':'
            << e.getColumnNumber () << ": ";

      switch (severity)
      {
      case Severity::warning:
        {
          wcerr << "warning: ";
          break;
        }
      default:
        {
          wcerr << "error: ";
          valid_ = false;
          break;
        }
      }

      wcerr << e.getMessage() << endl;
    }

  private:
    Boolean& valid_;
  };
  */


  // Failed to open resource.
  //
  struct Open {};

  class InputSource: public Xerces::InputSource
  {
  public:
    InputSource (
      Path const& abs,
      Path const& rel,
      Path const& base,
      Context const& ctx,
      Xerces::MemoryManager* mm = Xerces::XMLPlatformUtils::fgMemoryManager)
        : Xerces::InputSource (mm),
          abs_ (abs),
          rel_ (rel),
          base_ (base),
          ctx_ (ctx)
    {
      setSystemId (XML::XMLChString (
                     String (abs_.native_file_string ())).c_str ());
    }

    virtual Xerces::BinInputStream*
    makeStream () const
    {
      using namespace Xerces;

      BinFileInputStream* is (
        new (getMemoryManager ())
        BinFileInputStream (getSystemId (), getMemoryManager ()));

      if (!is->getIsOpen ())
      {
        delete is;

        wcerr << ctx_.file (base_) << ": error: "
              << "'" << rel_ << "': unable to open in read mode"
              << endl;

        throw Open ();
      }

      return is;
    }

  private:
    Path abs_;
    Path rel_;
    Path base_;
    Context const& ctx_;
  };


  class EntityResolver: public Xerces::DOMEntityResolver,
                        public Xerces::XMemory
  {
  public:
    EntityResolver (Context& ctx)
        : ctx_ (ctx)
    {
    }

    virtual Xerces::DOMInputSource*
    resolveEntity (XMLCh const* const pub_id,
                   XMLCh const* const prv_id,
                   XMLCh const* const base_uri)
    {
      /*
      XMLCh empty[1];
      empty[0] = 0;

      wcerr << "resolve entity:" << endl
            << "  pub_id " << (pub_id ? pub_id : empty) << endl
            << "  prv_id " << (prv_id ? prv_id : empty) << endl
            << "     uri " << (base_uri ? base_uri : empty) << endl;
      */

      // base_uri should be a valid path by now.
      //
      Path base (XML::transcode_to_narrow (base_uri),
                 boost::filesystem::native);

      if (prv_id == 0)
      {
        //@@ How can I get the line/column numbers for this?
        //
        wcerr << ctx_.file (base) << ": error: "
              << "unable to guess which resource to open"
              << endl;

        wcerr << ctx_.file (base) << ": info: "
              << "did you forget to specify schemaLocation for import/include?"
              << endl;

        throw Open ();
      }

      try
      {
        Path path (XML::transcode_to_narrow (prv_id));
        Path base_dir (base.branch_path ());

        Path abs_path, rel_path;

        if (path.is_complete ())
        {
          abs_path = rel_path = path;
        }
        else
        {
          abs_path = base_dir / path;
          rel_path = ctx_.file (base).branch_path () / path;
        }


        abs_path.normalize ();

        ctx_.map_file (abs_path, rel_path);

        using namespace Xerces;

        InputSource* is (
          new (XMLPlatformUtils::fgMemoryManager)
          InputSource (abs_path, rel_path, base, ctx_));

        // Note that I can't use XMLPlatformUtils::fgMemoryManager here
        // since Wrapper4InputSource is-not-an XMemory.
        //
        return new Wrapper4InputSource (is);
      }
      catch (InvalidPath const&)
      {
        wcerr << ctx_.file (base)
              << ": error: '" << prv_id << "' is not a valid filesystem path"
              << endl;
        throw;
      }

      // Will never reach.
      //
      return 0;
    }

  private:
    Context& ctx_;
  };



  XML::AutoPtr<Xerces::DOMDocument> Parser::Impl::
  dom (Path const& tu)
  {
    using namespace Xerces;

    try
    {
      // Initialize Xerces runtime
      //
      XMLPlatformUtils::Initialize();

      Context ctx;

      Path abs_path (system_complete (tu));
      abs_path.normalize ();

      ctx.map_file (abs_path, tu);

      InputSource input_source (abs_path, tu, abs_path, ctx);

      // First validate the schema with Xerces.
      //
      {
        // Instantiate the DOM parser.
        //
        XMLCh const gLS[] = {chLatin_L, chLatin_S, chNull };

        // Get an implementation of the Load-Store (LS) interface.
        //
        DOMImplementationLS* impl (
          static_cast<DOMImplementationLS*> (
            DOMImplementationRegistry::getDOMImplementation (gLS)));

        // Create a DOMBuilder.
        //
        XML::AutoPtr<DOMBuilder> parser (
          impl->createDOMBuilder (DOMImplementationLS::MODE_SYNCHRONOUS, 0));

        parser->setFeature (XMLUni::fgDOMComments, false);
        parser->setFeature (XMLUni::fgDOMDatatypeNormalization, true);
        parser->setFeature (XMLUni::fgDOMEntities, false);
        parser->setFeature (XMLUni::fgDOMNamespaces, true);
        parser->setFeature (XMLUni::fgDOMValidation, true);
        parser->setFeature (XMLUni::fgDOMWhitespaceInElementContent, false);
        parser->setFeature (XMLUni::fgXercesSchema, true);
        parser->setFeature (XMLUni::fgXercesSchemaFullChecking, true);
        parser->setFeature (XMLUni::fgXercesValidationErrorAsFatal, true);

        ErrorHandler eh (valid_, ctx);
        parser->setErrorHandler (&eh);

        EntityResolver er (ctx);
        parser->setEntityResolver (&er);

        Wrapper4InputSource wrap (&input_source, false);

        parser->loadGrammar (wrap, Grammar::SchemaGrammarType);
      }

      if (!valid_)
        return XML::AutoPtr<DOMDocument> (0);

      // Now do our own parsing.
      //

      std::auto_ptr<XSDDOMParser> xsd_parser (
        new (XMLPlatformUtils::fgMemoryManager) XSDDOMParser ());

      xsd_parser->setValidationScheme (XercesDOMParser::Val_Never);
      xsd_parser->setDoNamespaces (true);
      //xsd_parser->setUserEntityHandler(fEntityHandler);
      //xsd_parser->setUserErrorReporter(fErrorReporter);

      xsd_parser->parse (input_source);

      XML::AutoPtr<DOMDocument> doc (xsd_parser->adoptDocument());

      return doc;
    }
    catch (Xerces::XMLException const& e)
    {
      wcerr << tu << ": ice: Xerces::XMLException: " << e.getMessage ()
            << endl;

      abort ();
    }
    catch (Xerces::DOMException const& e)
    {
      Size const size = 2047;
      XMLCh text[size + 1];

      wcerr << tu << ": ice: Xerces::DOMException: ";

      if (DOMImplementation::loadDOMExceptionMsg (e.code, text, size))
        wcerr << text << endl;
      else
        wcerr << "no message available, error code: " << e.code << endl;

      abort ();
    }
    catch (InvalidPath const&)
    {
      // Diagnostics has already been issued.
      //
      valid_ = false;
    }
    catch (Open const&)
    {
      // Diagnostics has already been issued.
      //
      valid_ = false;
    }

    return XML::AutoPtr<DOMDocument> (0);
  }


  // Parser
  //
  Parser::
  ~Parser ()
  {
  }

  Parser::
  Parser ()
      : impl_ (new Impl)
  {
  }

  Evptr<SemanticGraph::Schema> Parser::
  parse (SemanticGraph::Path const& schema)
  {
    return impl_->parse (schema);
  }
}
